local gfind=string.gmatch or string.gfind;
DreamChat.messageTypes={}
DreamChat.messageCategs={}
DreamChat.slashes={}
DreamChat.languages={
	"COMMON",
	"DARNASSIAN",
	"DWARVISH",
	"GNOMISH",
	"DRAENEI",
	"ORCISH",
	"TAURAHE",
	"THALASSIAN",
	"GUTTERSPEAK",
	"TROLL",
	"DEMONIC",
	"DRACONIC",
	"KALIMAG",
	"TITAN",
	}
-- Todo: Localize/get from some WoW api?
DreamChat.languageNames={
	COMMON="Common",
	DARNASSIAN="Darnassian",
	DWARVISH="Dwarvish",
	GNOMISH="Gnomish",
	ORCISH="Orcish",
	TAURAHE="Taurahe",
	TROLL="Troll",
	GUTTERSPEAK="Gutterspeak",
	DEMONIC="Demonic",
	DRACONIC="Draconic",
	KALIMAG="Kalimag",
	TITAN="Titan",
	DRAENEI="Draenei",
	THALASSIAN="Thalassian",
}
DreamChat.languageTranslators={}
	
DREAMCHAT_FLAG_NOBRACKETS=1

-- @desc Registers a message type to be used later in DreamChat windows.
-- @arg typeId string The internal type identifier.
-- @arg visibleName string The default user-visible name for the chat type. Should localize.
-- @arg defaultSetting table? An optional table of settings (for future) for the message type.
-- @ret mixed True if succeeds, nil otherwise.
-- @cat Public
function DreamChat:RegisterMessageType(typeId,visibleName,defaultSettings,senderFunc)
	typeId=string.lower(typeId); -- Cuts through some ambiguity.
	if DreamChat.messageTypes[typeId]~=nil then
		DreamChat:Debug("Multiple registration for message type "..typeId..", ignored.")
		return nil
	end
	DreamChat.messageTypes[typeId]={id=typeId,name=visibleName,sender=senderFunc}
	if defaultSettings~=nil then
		local key,value;
		for key,value in pairs(defaultSettings) do
			DreamChat.messageTypes[typeId][key]=value;
		end
	end
	self:RebuildCategorization()
	if defaultSettings.slash~=nil then
		self:RebuildSlashes()
	end
	return true
end

-- @desc Unregisters a previously registered message type.
-- @arg typeId string The internal type identifier.
-- @ret mixed True if succeeds, nil otherwise.
-- @cat Public
function DreamChat:UnregisterMessageType(typeId)
	typeId=string.lower(typeId); -- Cuts through some ambiguity. Again.
	if DreamChat.messageTypes[typeId]==nil then
		DreamChat:Debug("Unregistration for unregistered message type "..typeId..", ignored.")
		return nil
	end
	DreamChat.messageTypes[typeId]=nil
	self:RebuildCategorization()
	return true
end

-- @desc Rebuilds the categorization table.
function DreamChat:RebuildCategorization()
	local k,v;
	DreamChat.messageCategs={}
	for k,v in pairs(DreamChat.messageTypes) do
		category=v.category or "(Other)";
		if DreamChat.messageCategs[category]==nil then
			DreamChat.messageCategs[category]={};
		end
		table.insert(DreamChat.messageCategs[category],k);
	end
	for k in pairs(DreamChat.messageCategs) do
		table.sort(DreamChat.messageCategs[k]);
	end
end

-- @desc Rebuilds the slash table.
function DreamChat:RebuildSlashes()
	local k,v,s;
	DreamChat.slashes={}
	
	for k,v in pairs(DreamChat.messageTypes) do
		if v.slash~=nil then
			for s in gfind(v.slash,"(%w+)") do
				if s~="" then
					DreamChat.slashes["/"..string.upper(s)]=k;
				end
			end
		end
	end
end

-- @desc Shows a message in DreamChat.
-- @arg typeId string A previously registered type ID. If the type ID has not been registered, no message will be shown.
-- @arg message string The message text.
-- @arg sender string? Optional string for message sender. This should be used whenever possible.
-- @arg status string? Optional string for status (<AFK>...)
-- @arg flags number? Combination of flags (see DREAMCHAT_FLAG_...)
-- @ret number Number of frames the message was output to or nil if error.
-- @cat Public
function DreamChat:ShowMessage(typeId,message)
	typeId=string.lower(typeId);
	if DreamChat.messageTypes[typeId]==nil then
		DreamChat:Debug("Inc message of type "..typeId.." ignored (not registered).")
		return nil
	end
	local frameId,frame,sent;
	sent=0;
	if type(message)=="string" then message={text=message}; end
	message.typeId=typeId;
	for frameId,frame in pairs(self.frames) do -- For each frame,
		if frame:AcceptsMessageType(typeId) then -- If they accept the message type,
			--DreamChat:Debug("Frame "..(frame.name).." accepts "..typeId..".")
			if frame:ShowMessage(message)~=nil then -- send it to the frame.
				sent=sent+1;
			end
		end
	end
	return sent;
end

-- @desc Sends a message via DreamChat.
-- @arg typeId string A previously registered type ID. If the type ID has not been registered, no message will be shown.
-- @arg message string The message text.
-- @arg flags number To be used later?
-- @ret boolean nil if error.
-- @cat Public
function DreamChat:SendMessage(typeId,message,flags)
	typeId=string.lower(typeId);
	if DreamChat.messageTypes[typeId]==nil then
		DreamChat:Debug("Out message of type "..typeId.." ignored (not registered).")
		return nil
	end
	if DreamChat.messageTypes[typeId].sender==nil then
		DreamChat:Debug("Out message of type "..typeId.." ignored (no sender function).")
		return nil
	end
	return DreamChat.messageTypes[typeId].sender(typeId,message,flags) or 1;
end

-- @desc Registers some WoW default message types and their event listeners.
function DreamChat:RegisterDefaultMessageTypes()
	-- Todo: Localize.
	-- Default types.
	self:RegisterMessageType("say","Say", {color="FFFFFF", category="Say", slash="say s", pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("yell","Yell", {color="FF3333", category="Say", slash="yell y",nonSticky=1, pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("guild","Guild", {color="EEFFEE",category="Guild",slash="guild g", pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("officer","Officer", {color="FFCC00",category="Guild",slash="officer off o", pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("party","Party", {color="EEEEFF",category="Party",slash="party p", pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("raid","Raid", {color="CC00FF",category="Party",slash="raid ra", pratRenderable=1},self.MessageSender) -- random color, meh
	self:RegisterMessageType("raid_leader","Raid Leader", {color="CC00FF",category="Party", pratRenderable=1})
	self:RegisterMessageType("raid_warning","Raid Warning", {color="CC00FF",category="Party",slash="rw raidw",nonSticky=1, pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("battleground","Battleground", {color="EE00FF",category="Party",slash="bg", pratRenderable=1},self.MessageSender)
	self:RegisterMessageType("battleground_leader","Battleground Leader", {color="EE00FF",category="Party", pratRenderable=1})
	self:RegisterMessageType("system","System",{color="00CCFF",category="System", pratRenderable=1},self.MessageSender)
	self:RegisterEvent("CHAT_MSG_SAY","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_YELL","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_PARTY","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_GUILD","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_OFFICER","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_RAID","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_RAID_LEADER","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_RAID_WARNING","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_BATTLEGROUND","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_BATTLEGROUND_LEADER","GenericMessageHandler")
	self:RegisterEvent("CHAT_MSG_SYSTEM","GenericMessageHandler")
	-- Slightly more esoteric types.
	self:RegisterMessageType("emote","Emote",{color="DDDDDD",category="Say",slash="me em e", pratRenderable=1},self.MessageSender)
	self:RegisterEvent("CHAT_MSG_EMOTE","EmoteHandler")
	self:RegisterEvent("CHAT_MSG_TEXT_EMOTE","EmoteHandler")
	-- Really kerrrazy types. That's basically channels and whispers.
	local i;
	for i=1,10 do
		self:RegisterMessageType("channel"..i,"Channel #"..i,{color="AFEEEE",category="Channel",slash=tostring(i), pratRenderable=1},self.MessageSender)
		self:RegisterMessageType("channelnotify"..i,"Channel #"..i,{color="AFEEEE",category="Channel Notifications",prefix="*"})
	end
	self:RegisterEvent("CHAT_MSG_CHANNEL","ChannelHandler")
	self:RegisterEvent("CHAT_MSG_CHANNEL_JOIN","ChannelNotifyHandler")
	self:RegisterEvent("CHAT_MSG_CHANNEL_LEAVE","ChannelNotifyHandler")
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE_USER","ChannelNotifyHandler")
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE","ChannelNotifyHandler")
	self:RegisterEvent("CHAT_MSG_CHANNEL_LIST","ChannelNotifyHandler")
	self:RegisterMessageType("whisper","Whisper",{color="DDA0DD",category="Whisper",slash="w",slashValidator=self.ValidateWhisper,headerText="To PARAM",nonSticky=1, pratRenderable=1},self.MessageSender) -- Fun factoid: That's the web color "plum".
	self:RegisterMessageType("whisperto","Outgoing Whisper",{color="DDA0DD",category="Whisper",prefix="To ",dontShowName=true, pratRenderable=1})
	self:RegisterEvent("CHAT_MSG_WHISPER","WhisperHandler")
	self:RegisterEvent("CHAT_MSG_WHISPER_INFORM","WhisperHandler")
	
	-- DreamChat system types.
	self:RegisterMessageType("print","Print",{color="B00B15",category="System",slash="pr print"},self.CodeMonkeySender)
	self:RegisterMessageType("script","Script",{color="B00B15",category="System",slash="z script"},self.CodeMonkeySender)
	self:RegisterMessageType("dc","DreamChat",{color="FF3300",category="System",dontShowName=true})
	-- Todo: Add the rest
end

-- @desc Does some initialization for the messaging core
function DreamChat:InitializeMessaging()
	self:RegisterDefaultMessageTypes()
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE","SetChannelNames")
	self:SetChannelNames()
	--self.whisperTarget=nil
	self.language=nil
end

-- @desc Sets the current output language.
-- @ret boolean 1 if succeeded (allowed language)
function DreamChat:SetLanguage(lang)
	local k,v;
	for k,v in ipairs(self.languages) do
		if v==lang then
			self.language=lang
			return 1
		end
	end
	return nil
end

-- @desc Adds a custom language translator.
-- @arg identifier string The language identifier. Will be passed to the translator func.
-- @arg displayedName string The displayed name for the language.
-- @arg translator string The translator function(identifier,message). to return the translated message.
-- @ret boolean 1 if succeeded, nil if not (the language already exists)
function DreamChat:AddCustomLanguage(identifier,displayedName,translator)
	identifier=string.upper(identifier);
	if self.languageNames[identifier]~=nil then
		self:Debug("The language \""..identifier.."\" already exists; won't create a custom language for it.")
		return nil
	end
	table.insert(self.languages,identifier)
	self.languageNames[identifier]=displayedName or identifier
	self.languageTranslators[identifier]=translator
	return 1
end

-- @desc Sets the channel names into channel message types when they change.
function DreamChat:SetChannelNames()
	for i=1,10 do
		local _,n=GetChannelName(i);
		n=string.format('%02d: %s',i,(n or "(none)"))
		self.messageTypes['channel'..i].name=n;
		self.messageTypes['channelnotify'..i].name=n;
	end
end




-- @desc The event handler function for say/yell/party/guild/officer/raid/bg messages.
-- @cat ChatHandler
function DreamChat:GenericMessageHandler(message,sender,language)
	local typeId=string.lower(strsub(event,10))
	--self:Debug("GMH: %s, %s, %s",message,sender,language)
	self:ShowMessage(typeId,{sender=sender,text=message,language=language})
end

-- @desc The event handler function for emotes.
-- @cat ChatHandler
function DreamChat:EmoteHandler(message,sender)
	if event=="CHAT_MSG_TEXT_EMOTE" then sender=nil end
	self:ShowMessage("emote",{sender=sender,text=message})
end

-- @desc The magical, insane channel handler.
-- @cat ChatHandler
function DreamChat:ChannelHandler(message,sender,language,channel,target,status,zone,chanNum,chanName)
	self:ShowMessage("channel"..chanNum,{channel=tonumber(chanNum),channelName=chanName,sender=sender,text=message,language=language,status=status})
	-- Todo: Add better magic.
end

-- @desc The channel notice handler.
-- @cat ChatHandler
function DreamChat:ChannelNotifyHandler()
	if event=="CHAT_MSG_CHANNEL_JOIN" then
		-- arg2 Name of the player that joined 
		-- arg4 Number and name of the channel (e.g. "5. MyOwnChannel") 
		-- arg8 Channel number 
		-- arg9 Window title
		self:ShowMessage("channelnotify"..arg8,{sender=arg2,text=" joined channel.",subType="join"})
	elseif event=="CHAT_MSG_CHANNEL_LEAVE" then
		-- arg2 Player name that left 
		-- arg7 Channel number 
		-- arg8 Channel name without number 
		self:ShowMessage("channelnotify"..arg7,{sender=arg2,text=" left channel.",subType="leave"})
	elseif event=="CHAT_MSG_CHANNEL_LIST" then
		-- arg1 The list of values displayed by ListChannels() or ListChannelByName() in one string. 
		-- arg4 The number and name of the channel the message came from. ie: "1. General"
		local _,cn=strfind(arg4,"(%d+). (.+)")
		if cn~=nil then
			self:ShowMessage("channelnotify"..cn,arg1)
		end
	elseif event=="CHAT_MSG_CHANNEL_NOTICE" then
		-- arg1 "YOU_JOINED" if you joined a channel, or "YOU_LEFT" if you left
		-- arg4 Channel name with number (e.g. "6. TestChannel") 
		-- arg8 Channel Number 
		-- arg9 Channel name without number
		local me=UnitName("player");
		if arg1=="YOU_JOINED" then
			self:ShowMessage("channelnotify"..arg8,{channel=arg8,sender=me,text=" joined channel.",subType="join"})
		else
			self:ShowMessage("channelnotify"..arg8,{channel=arg8,sender=me,text=" left channel.",subType="leave"})
		end
	elseif event=="CHAT_MSG_CHANNEL_NOTICE_USER" then
		-- arg1 "ANNOUNCEMENTS_OFF", "ANNOUNCEMENTS_ON", "BANNED", "CHANNEL_OWNER", "INVALID_NAME", "INVITE", "MODERATION_OFF", "MODERATION_ON", "MUTED", "NOT_MEMBER", "NOT_MODERATED"
		-- arg2 If arg5 has a value then this is the user effected ( eg: "Player Foo has been kicked by Bar" ), if arg5 has no value then it's the person who caused the event ( eg: "Channel Moderation has been enabled by Bar" ) 
		-- arg4 Channel name with number 
		-- arg5 Player that caused the event (eg "Player Foo has been kicked by Bar" )
		local _,cn=strfind(arg4,"(%d+). (.+)")
		if cn~=nil then
			self:ShowMessage("channelnotify"..cn,{channel=cn,sender=arg5,text=arg2.." "..arg1})
		end		
	end
end

-- @desc The whisper handler.
-- @cat ChatHandler
function DreamChat:WhisperHandler(message,sender,status)
	if event=="CHAT_MSG_WHISPER_INFORM" then
		self:ShowMessage("whisperto",{sender=sender,text=message})
	else
		self:ShowMessage("whisper",{sender=sender,text=message,status=statusOrLang})
	end
	-- Todo: Revise.
end

DreamChat.MessageSenderMap=
{
    whisper="WHISPER",
    whisperto="WHISPER_INFORM",
    channel1="CHANNEL",
    channel2="CHANNEL",
    channel3="CHANNEL",
    channel4="CHANNEL",
    channel5="CHANNEL",
    channel6="CHANNEL",
    channel7="CHANNEL",
    channel8="CHANNEL",
    channel9="CHANNEL",
    channel10="CHANNEL",
    
	say="SAY",
	yell="YELL",
	party="PARTY",
	raid="RAID",
	raid_warning="RAID_WARNING",
	battleground="BATTLEGROUND",
	guild="GUILD",
	officer="OFFICER",
	emote="EMOTE"
	
	
}

-- @desc The generic message sender.
-- @cat ChatSender NonMethod
-- @note THIS FUNCTION IS NOT A DREAMCHAT METHOD, BUT A FUNCTION DEFINED IN THE DREAMCHAT TABLE.
function DreamChat.MessageSender(typeId,message,param)
	if DreamChat.language~=nil and DreamChat.languageTranslators[DreamChat.language]~=nil then
		local nmessage=DreamChat.languageTranslators[DreamChat.language](DreamChat.language,message);
		if nmessage~=nil then message=nmessage; end
	end
	if DreamChat.MessageSenderMap[typeId] then
		return SendChatMessage(message,DreamChat.MessageSenderMap[typeId],string.upper(DreamChat.language))
	end
	if typeId=="whisper" then
		DreamChat.lastWhispered=DreamChat.editBoxParam;
		return SendChatMessage(message,"WHISPER",string.upper(DreamChat.language),param or DreamChat.editBoxParam)
	end
	if string.sub(typeId,1,7)=="channel" then
		return SendChatMessage(message,"CHANNEL",string.upper(DreamChat.language),string.sub(typeId,8))
	end
	DreamChat:ShowMessage("dc","trying to send a message of type <"..typeId.."> through the generic sender, no dice")
end

-- @desc Validates a whisper command. Internally, this sets the correct whisper target and returns the editbox header string.
-- @note THIS FUNCTION IS NOT A DREAMCHAT METHOD, BUT A FUNCTION DEFINED IN THE DREAMCHAT TABLE.
-- @cat NonMethod
function DreamChat.ValidateWhisper(text)
	local a,b,name,msg=string.find(text,"^/w ([^ ]-) (.*)")
	if a==nil or string.len(name)<2 then
		DreamChat.editBoxParam=nil
		return nil
	end
	DreamChat.editBoxParam=name
	return msg
end

-- @desc Debugging sender
-- @note THIS FUNCTION IS NOT A DREAMCHAT METHOD, BUT A FUNCTION DEFINED IN THE DREAMCHAT TABLE.
-- @cat ChatSender NonMethod
function DreamChat.CodeMonkeySender(typeId,message,param)
	if typeId=="print" then
		RunScript("DCCM_temp=("..message..");");
		DreamChat:ShowMessage("print",tostring(DCCM_temp),message);
		DCCM_temp=nil;
	end
	if typeId=="script" then
		RunScript(message);
	end
end

-- @desc Gets a message type color triplet for the given typeid, if any
-- @ret r,g,b number,number,number the values
function DreamChat:GetMessageTypeColor(typeId)
	if typeId==nil then
		return 1,0,0.5;
	elseif DreamChat.db.profile.colors[typeId]~=nil then
		return unpack(DreamChat.db.profile.colors[typeId]);
	elseif DreamChat.messageTypes[typeId].color then
		return DreamChat:HexToRGB(DreamChat.messageTypes[typeId].color);
	else
		return 1,1,1;
	end
end
